// Copyright 2025 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package fusetesting

import (
	"fmt"
	"os"
	"sort"
)

// Read the directory with the given name and returns a list of os.FileInfo entries,
// sorted by name.
func ReadDirPlusPicky(dirname string) (entries []os.FileInfo, err error) {
	// Open the directory.
	f, err := os.Open(dirname)
	if err != nil {
		return nil, fmt.Errorf("Open: %v", err)
	}

	// Don't forget to close it later.
	defer func() {
		closeErr := f.Close()
		if closeErr != nil && err == nil {
			err = fmt.Errorf("Close: %v", closeErr)
		}
	}()

	// Read all directory entries and their FileInfo.
	entries, err = f.Readdir(-1)
	if err != nil {
		return nil, fmt.Errorf("Readdir: %v", err)
	}

	// Sort the collected FileInfo entries by their name.
	sort.Sort(sortedEntries(entries))

	return entries, nil
}
